#Param: (been) simple routine for command line argument / config file processing

#Copyright (c) 2004-2008 Thomas Orgis <thomas@orgis.org>
#This is Open Source, Distribution and Modification under the terms of the Artistic License

# still should break Param::Parse() down, these 300 lines are un-understandable for an outsider

# current modus operandi:
# 1. parse command line args into temporary storage
# 2. parse config files (that may have been set on command line)
# 3. merge settings (overwrite config file settings)

# I need to retain the behaviour... so in the case of special operators I have to store the operations that happen on command line
# ...do all config file parsing... and then apply the command line
# the normal command line is para=xx parb=zz para=xz ...
# instead of a list of settings I have to make a list of operations
# %ops{key} = ['=', 'stuff', '.=', 'more stuff'];

package Param;

use strict;
# What special magic is happening here compared to declaring the var with my/our/... ?
use vars qw($VERSION $version %features);

# an attempt to comply with Perl module version convention
# I should check how that relates to the version compare logic I have in mind.
$VERSION = '2.25_0';
# a nicer string
$version = $VERSION.' (19.04.2008)';
$version =~ s/_/./;
%features = qw(array 1);

use Sys::Hostname;
use FindBin qw($Bin);
use File::Spec;
use File::Basename;

#command line args parsing

#i predefine param{help} with switch h!!!
#i predefine param{config} with switch I!!!

my $ops = '.+\-*\/';
my $sopex = '['.$ops.']?=|[.+\-*\/]';
my $lopex = '['.$ops.']?=';

sub Parse #(config, params)
{
	my $me = '[Param::Parse]';
	my $noop = '[^+\-=.*\/]'; # needed for parsing and option name validity check
	my $config = {};
	# will make this better!
	if(not defined $_[0] or $_[0] =~ '^HASH'){ $config = shift }
	$config->{program} = basename($0) unless defined $config->{program};
	
	my $printconfig = 0;
	#$config->{verbose} = 1;
	my $hh = 'show the help message, give value > 1 for more help';
	my $ih = 'Which configfile to use;'."\n".'special: just -I or --config causes printing a current config file to STDOUT';
	# todo: stuff these into less (one) data structures
	my %param = ('help',0, 'config', undef);	
	my %parexist = ('help',1,'config',1); #for checking for unknown parameters
	my %par_help = ('help',\$hh,'config',\$ih);
	my %par_map = ('h','help','I','config');
	my %short = ('help','h','config','I');
	my %par_array = (); # marker for parameters holding a list

	#new 2.15 feature: version info
	if(defined $config->{version})
	{
		$param{version} = 0;
		$par_help{version} = \'print out the program version';
		$parexist{version} = 1;
		$short{version} = '';
	}
	
	my $length = 6;
	$param{config} = $config->{file};

	#determine directory to search config files in
	my $confdir = $config->{confdir};
	unless(defined $confdir)
	{
		foreach my $d
		(
			File::Spec->catfile($ENV{HOME},'.'.$config->{program})
			,File::Spec->catfile($Bin,'..','etc',$config->{program})
			,File::Spec->catfile($Bin,'..','etc')
			,File::Spec->catfile($Bin,'etc')
			,$Bin
		)
		{
			if(-d $d){ $confdir = $d; last; }
		}
	}
	print STDERR "Looking for config files in $confdir\n" if $config->{verbose};
	#means: nofile[false,true], file[string], info, verbose[bool],  
	#config confusion
	unless($config->{nofile} or $param{config}) #as long as I was told not to use a config file or it has been already given
	{
		#scan for a config file
		my @l = ( File::Spec->catfile($confdir,$config->{program}.'.conf'), File::Spec->catfile($confdir,$config->{program}.'.'.hostname().'.conf') );
		print STDERR "possible config files: @l\n" if $config->{verbose};
		if($config->{multi})
		{
			#well, this is a nice example of totally relying on a two-valued array
			shift(@l) unless (-f $l[0]); #remove default file from list when not existing
			pop(@l) unless (-f $l[$#l]); #remove hostfile from list when not existing
		}
		else
		{
			pop(@l) unless (-f $l[$#l]); #only when existing
			shift(@l) if ($#l == 1 or ! (-f $l[0])); #hehe...
		}
		$param{config} = join(',', @l); #to be compatible with command line input
	}
	#create parameter entries
	my $pars = [];
	my $args = [];
	if($#_ > 1 or $#_ == -1)
	{
		$pars = \@_;
		$args = \@ARGV;
	}
	else
	{
		$pars = shift;
		$args = defined $_[0] ? shift : \@ARGV;
	}
	my $r;
	my $t;
	die "Param: Wrong Parameter count. This is the Programmer's fault.\n" if (($#{$pars}+1) % 4) != 0;
	
	if($config->{gimme} and $#{$args}==-1){ $param{help} = 1} #giving help when no arguments
	
	while(defined ($r = shift(@{$pars})))
	{
		my ($val,$short,$help) = (shift(@{$pars}),shift(@{$pars}),\(shift(@{$pars})));
		# check the definition for validity, don't have (too many) operational characters in names
		unless( ($short eq '' or $short =~ /^$noop$/o) and not $r =~ /=/ and $r =~ /^$noop/o and $r =~ /$noop$/ )
		{
			print STDERR "Param: Invalid parameter (long/short) name $r/$short!\nProgrammer, please fix this (sleeping for a second to annoy you).\n";
			sleep(1);
			next;
		}
		unless(defined $param{$r} or defined $par_map{$short})
		{
			$length = length($r) if length($r) > $length;
			$param{$r} = $val;
			$par_array{$r} = 1 if("$val" =~ '^ARRAY');
			$parexist{$r} = 1;
			$par_map{$short} = $r if $short ne '';
			$par_help{$r} = $help;
			$short{$r} = $short;
		}
		else{print STDERR "NewPar: Tried to redefine an option ($r / $short)! Programmer: please check this!\n"; return -1;}
	}
	
	#now interpret command line opts / config file
	my %ops = ();

	#regexes for option parsing
	#the companion script paramex gives a picture...
	my $shorts = '^(([-+])('.$noop.'+|('.$noop.')('.$sopex.')(.*)))$'; #allow -x -x=bla -xyz
	my $longs = '^(([-+]{2})([^=]{2,})(('.$lopex.')(.*)|))$'; #allow --long --long=lala
	my $separator = '^--$'; #exactly this string means "Stop the parsing!"

	if($config->{lazy})
	{
		$shorts = '^(([-+]?)('.$noop.'+|('.$noop.')('.$sopex.')(.*)))$'; #allows -x -x=bla x x=bla xyz -xyz
		$longs = '^(([-+]{0,2})([^=]{2}[^=]*)()('.$lopex.')(.*)|(--|\+\+)([^=]+))$';
		$separator = '^-+$'; # allows --long=val -long=val long=val --long
	}

	# the option parser, code for short and long options is separated, but mostly similar
	while(defined $args->[0])
	{
		print STDERR "[Param::Parse] parsing $args->[0]\n" if $config->{verbose};
		my $e = index($args->[0], "\n");
		my $begin;
		my $end = "";
		if($e >=0)
		{
			$begin = substr($args->[0],0,$e);
			$end   = substr($args->[0],$e);
		}
		else
		{
			$begin = $args->[0];
		}
		if($begin =~ /$separator/o)
		{
			shift(@{$args});
			last;
		}
		elsif($begin =~ /$shorts/o)
		{
			my $sign = $2;
			$sign = '-' if $sign eq '';
			my $sname = $4;
			my $op = $5;
			my $val = (defined $6 ? $6 : '').$end;
			unless(defined $op)
			{
				$sname = $3;
				$val = $sign =~ /^-/ ? 1 : 0;
				$op = '=';
				if($sname eq 'I'){ $printconfig += 1; undef $op; }
			}
			if(defined $op)
			{
				$op .= '=' unless $op =~ /=$/;
				my @names; # the list of real parameters
				# check for valid parameter names
				foreach my $s (split(//,$sname))
				{
					if(defined (my $name = $par_map{$s}))
					{
						push(@names, $name);
					}
					else
					{
						if($config->{fuzzy})
						{
							print STDERR "$me Unknown short option $s, assuming that this is data instead.\n" if $config->{verbose};
							@names = ();
							last;
						}
						else
						{
							$param{help} = 1;
							&ParError($s,\%par_map);
						}
					}
				}
				last unless @names; # end parsing if switches invalidated

				foreach my $n (@names)
				{
					print STDERR "[Param::Parse] param $n\n" if $config->{verbose};
					if($op eq '='){ $ops{$n} = [ '=', $val ]; }
					else{ push(@{$ops{$n}}, $op, $val); }
				}
			}
			shift(@{$args});
		}
		elsif($begin =~ /$longs/o)
		{
			#yeah, long option
			my $sign = defined $7 ? $7 : $2;
			$sign = '--' if $sign eq '';
			my $name = defined $8 ? $8 : $3;
			print STDERR "[Param::Parse] param $name\n" if $config->{verbose};
			my $op = $5;
			my $val = (defined $6 ? $6 : '').$end;
			print STDERR "[Param::Parse] val $val\n" if $config->{verbose};
			print STDERR "[Param::Parse] from $begin$end\n" if $config->{verbose};
			# hack for operators, regex may swallow the . in .=
			unless($name =~ /$noop$/o)
			{
				$op = substr($name,-1,1).$op;
				$name = substr($name,0,length($name)-1);
			}
			unless(defined $op)
			{
				$val = $sign =~ /^-/ ? 1 : 0;
				$op = '=';
				if($name eq 'config'){ $printconfig += 1; undef $op; }
			}
			if(defined $op)
			{
				if($parexist{$name})
				{
					$op .= '=' unless $op =~ /=$/;
					if($op eq '='){ $ops{$name} = [ '=', $val ]; }
					else{ push(@{$ops{$name}}, $op, $val); }
				}
				else
				{
					if($config->{fuzzy})
					{
						print STDERR "$me Unknown long option $name, assuming that this is data instead.\n" if $config->{verbose};
						last;
					}
					else
					{
						$param{help} = 1; 
						&ParError($name,\%param);
					}
				}
			}
			shift(@{$args});
		}
		else{ last; } #was no option... consider the switch part over
	}

	# do operations on config file parameter first
	if(defined $ops{config})
	{
		while(@{$ops{config}})
		{
			my $op = shift(@{$ops{config}});
			my $val = shift(@{$ops{config}});
			eval('$param{config}'.$op.'$val');
		}
		delete $ops{config};
	}

	#now parse config file
	if($param{config} ne "")
	{
		ReadFile($confdir, $config, \%param, \%parexist, undef, undef, \%par_array);
	}
	
	#and now apply the explicit command line switches,options,parameters,...
	foreach my $key (keys %ops)
	{
		print STDERR "Param: applying (@{$ops{$key}}) to $key\n" if $config->{verbose};
		while(@{$ops{$key}})
		{
			my $op = shift(@{$ops{$key}});
			my $val = shift(@{$ops{$key}});
			apply_op(\$param{$key}, \$op, \$val, $par_array{$key});
		}
	}

	unless($config->{nofinals})
	{

		#give the help (info text + option help) and exit when -h or --help was given
		if($param{help} and not $param{ignorehelp}) #ignorehelp to vanish in 3.0 (at least become complete alias of nofinals)
		{ # ugly... need to consolidate parameters ... want to avoid real object, though.
			Help($config, \%param, \%short, \%par_help, \%par_array, $length);
			exit(0);
		}

		#give version info
		elsif(defined $config->{version} and $param{version})
		{
			print "$config->{program} $config->{version}\n";
			exit(0);
		}

		elsif($printconfig)
		{
			PrintFile(\*STDOUT, $config, \%param, \%short, \%par_help, 0, ($printconfig > 1), \%par_array);
			exit(0);
		}

	}

	return $config->{returnall} ? (\%param,  \%short, \%par_help, \%parexist) : \%param;	
}

sub Help
{
	my ($config, $param, $short, $par_help, $par_array, $length) = @_;
	my %example = 
	(
		 'lazy'   => "[-]s xyz [-]s=value -[-]long -[-]long=value - <files/stuff>\n"
		,'normal' => "-s -xyz -s=value --long --long=value [--] <files/stuff>\n"
	);

	print "\n$config->{program} ".(defined $config->{version} ? "v$config->{version} " : '')."- ",$config->{info},"\n\n" if defined $config->{info};
	print "Generic parameter example (list of real parameters follows):\n\t$config->{program} ";
	if($config->{lazy})
	{
		print $example{lazy};
		if($param->{help} > 1)
		{
			print "The [ ] notation means that the enclosed - is optional, saving typing time for really lazy people. Note that \"xyz\" mentions three short options, opposed to the long option \"-long\". In trade for the shortage of \"-\", the separator for additional unnamed parameters is mandatory (supply as many \"-\" grouped together as you like;-).\n";
		}
	}
	else
	{
		print $example{normal};
	}
	if($param->{help} > 1)
	{
		print "\n";
		print "You mention the parameters/switches you want to change in any order or even multiple times (they are processed in the oder given, later operations overriding/extending earlier settings.\n";
		print "An only mentioned short/long name (no =value) means setting to 1, which is true in the logical sense. Also, prepending + instead of the usual - negates this, setting the value to 0 (false).\n";
		print "Specifying -s and --long is the same as -s=1 and --long=1, while +s and ++long is the sames as -s=0 and --long=0.\n";
		print 'There are also different operators than just "=" available, notably ".=", "+=", "-=", "*=" and "/=" for concatenation / appending array element and scalar arithmetic operations on the value.'."\n";
	}
	else
	{ # Don't waste so many lines by default.
		print "Just mentioning -s equals -s=1 (true), while +s equals -s=0 (false).\n";
		print "Using separator \"--\" makes sure that parameter parsing stops.\n"
	}
	print "\n";
	print "Recognized parameters:\n";
	print "NAME, SHORT [VALUE] DESCRIPTION\n";
	my @hidden;
	# TODO: adapt to current terminal width with proper line breaks.
	foreach my $k ( sort keys %{$param} )
	{
		unless($config->{hidenonshort} and $short->{$k} eq '' and not ($k eq 'version' and defined $config->{version}))
		{
			my @help = split("\n",${$par_help->{$k}});
			my $content = defined $param->{$k}
				? '['.($par_array->{$k}
					 ? ( @{$param->{$k}} ? "'".join("','", @{$param->{$k}})."'" : '' )
					 : $param->{$k}).']'
				: '<undef>';
			$help[0] = $content.' '.$help[0];
			print "$k", $short->{$k} ne '' ? ", $short->{$k} " : "    ";
			my $i = length($k) + 4;
			foreach my $h (@help)
			{
				for(; $i < $length + 4; ++$i){ print " "; }
				$i = 0;
				print "$h\n";
			}
		}
		else{ push(@hidden, $k); }
	}
	if($param->{help} > 1 and @hidden)
	{
		print "\nThere ".($#hidden == 0 ? 'is one hidden parameter' : "are ".($#hidden+1)." hidden parameters")." intended for use in config files (but also settable via cmdline):\n\t@hidden\nRun me with just -I or --config to get a config file on stdout (the console) with the help info and current values.\n";
	}
	elsif(@hidden)
	{
		print "\nHidden parameters: @hidden\n";
	}
	print "\n";
}

sub ParError
{
	my $par = shift;
	my $hash = shift;
	print STDERR "Unknown parameter/switch \"$par\" not in (",join(',',keys %{$hash}),")!\n";
}

sub ParseFile
{
	my $config = shift;
	my %param;
	ReadFile('.', $config, \%param);
	return \%param;
}

sub ReadFile
{
	# OK, OK, I should move this into OO design...
	my ($confdir,$config, $param, $parexist, $short, $parhelp, $par_array) = @_;
	my $parsing = (defined $short and defined $parhelp) ? 1 : 0;
	my $curkey = undef;
	$par_array = {} unless defined $par_array;

	my $lend = '\012\015';
	my $unspecial = '[^\#'.$ops.'=\s<\012\015]';

	my @l;
	if(defined $param->{config})
	{
		@l = ($config->{multi}) ? split(',',$param->{config}) : ($param->{config});
	}
	else{ @l = ($config->{file}) if defined $config->{file}; }
	foreach my $file (@l)
	{
		require IO::File;
		$file =~ s/^~\//$ENV{HOME}\// if defined $ENV{HOME};
		if(not -f $file and -f $file.'.conf'){ $file .= '.conf'; }
		$file = File::Spec->catfile($confdir,$file) unless (-f $file); #look in confidir
		if(not -f $file and -f $file.'.conf'){ $file .= '.conf'; }
		print STDERR "Parsing $file...\n" if $config->{verbose};
		my $cdat = new IO::File;
		if($cdat->open($file))
		{
			if(defined $config->{binmode})
			{
				binmode($cdat, $config->{binmode});
			}
			#do we need or want binmode for newlines?
			my $multiline = '';
			my $mcollect = 0;
			my $ender = '';
			my $mkey ='';
			my $mop = '';
			while(<$cdat>)
			{
				unless($mcollect)
				{
					unless($_ =~ /^\s*#?\s*[$lend]*$/)
					{
						if($parsing and $_ =~ /^=([^$lend]+)[$lend]*$/)
						{
							my $meta = $1;
							if($meta =~ /^param file\s*(\(([^)]*)\)|)\s*for\s*(.+)$/)
							{
								print STDERR "This file is for $3.\n" if $config->{verbose};
								$config->{program} = $3;
								if(defined $2 and $2 =~ /^(.+)$/)
								{
									foreach my $s (split(',', $1))
									{
										print STDERR "Activating option $s.\n" if $config->{verbose};
										$config->{$s} = 1;
									}
								}
							}
							elsif($meta =~ /^key\s*(\S*)(\s*short\s*(\S)|)/)
							{
								$curkey = $1;
								$short->{$curkey} = defined $3 ? $3 : '';
								${$parhelp->{$1}} = '';
								$param->{$1} = undef;
								$parexist->{$1} = 1;
								print STDERR "switching to key $1 / $3\n" if $config->{verbose};
							}
							elsif($meta =~ /^desc\s(.*)$/)
							{
								${$parhelp->{$curkey}} .= ${$parhelp->{$curkey}} ne '' ? "\n" : '' . $1 if defined $curkey;
							}
							elsif($meta =~ /^info\s(.*)$/)
							{
								$config->{info} .= $1."\n"; #dos, unix... whatever...
							}
							elsif($meta =~ /^version\s*(.+)$/)
							{
								$config->{version} = $1;
							}
						}
						else
						{
							if($_ =~ /^\s*($unspecial+)\s*($lopex)\s*([^$lend]*)[$lend]*$/)
							{
								my ($par,$op,$val) = ($1,$2,$3);
								#remove trailing spaces
								$val =~ s/\s*$//;
								#remove quotes
								$val =~ s/^"(.*)"$/$1/;
								if(not defined $parexist or $parexist->{$par})
								{
									print STDERR "Setting $par $op $val ", ($config->{eval} ? "(evaluating first)" : ""), "\n" if $config->{verbose};
									$val = eval($val) if ($config->{eval} and not $parsing);
									apply_op(\$param->{$par}, \$op, \$val, $par_array->{$par});
									# sense? perhaps we _wanted_ undef
									#$param->{$par} = '' unless defined $param->{$par};
								}
								else
								{
									if($config->{nanny}){ $param->{help} = 1; }
									print STDERR "$config->{program}: unknown parameter $par!\n";
								}
							}
							elsif($_ =~ /^\s*($unspecial+)\s*[$lend]*$/)
							{
								if(not defined $parexist or $parexist->{$1})
								{
									print STDERR "Setting $1 to 1\n" if $config->{verbose}; 
									$param->{$1} = $par_array->{$mkey} ? [ 1 ] : 1;
								}
								else
								{
									if($config->{nanny}){ $param->{help} = 1; }
									print STDERR "$config->{program}: unknown parameter $1!\n";
								}
							}
							elsif($_ =~ /^\s*($unspecial+)\s*([$ops]?)<<(.*)$/)
							{
								$ender = $3;
								$mop = $2 ne '' ? $2 : '=';
								$mkey = $1;
								$mcollect = 1;
								$mop .= '=' unless $mop =~ /=$/;
								print STDERR "Reading for $mkey...(till $ender)\n" if $config->{verbose}; 
							}
						}
					}
				}
				else
				{
					print STDERR "collect: $_" if $config->{verbose};
					unless($_ =~ /^$ender$/)
					{
						s/[$lend]+$/\n/;
						$multiline .= $_;
					}
					else
					{
						$mcollect = 0;
						if(not defined $parexist or $parexist->{$mkey})
						{
							my $val = ($config->{eval} and not $parsing) ? eval($multiline) : $multiline; #apply the config file options first, with eval() when desired
							#$param->{$mkey} = '' unless defined $param->{$mkey}; #really sensible here?
							apply_op(\$param->{$mkey}, \$mop, \$val, $par_array->{$mkey});
							print STDERR "set $mkey from $multiline\n" if $config->{verbose};
						}
						else
						{
							if($config->{nanny}){ $param->{help} = 1; }
							print STDERR "$config->{program}: unknown parameter $mkey!\n";
						}

						$multiline = '';
					}
				}
			}
			$cdat->close();
			print STDERR "...done parsing.\n" if $config->{verbose};
		}
		else{print STDERR "Couldn't open config file $file! ($!)\n" unless $config->{nocomplain};}
	}
}


#(
sub PrintFile
{
	my ($handle, $config, $param, $short, $par_help, $raw, $bare, $is_array) = @_;
	my @omit = ('config','help');
	push(@omit,'version') if defined $config->{version};
	push(@omit, @{$config->{notinfile}}) if defined $config->{notinfile};
	$is_array = {} unless defined $is_array;
	unless($bare)
	{
		print $handle <<EOT;
# Configuration file for $config->{program}
#
# Syntax:
# 
EOT
	if($config->{eval})
	{
		print $handle <<EOT;
# 	name = perl expression
#
# Enclosing "" are optional and you can provide any number (including 0) of whitespaces before and after name and the expression.
# The expression gets eval()ed in-place and the result is stored in the parameter named name. When you want to set a plain string, use double quoting (""string"" or "'string'", when you know Perl, you know the difference;-) or just single single quotes ('value').
EOT
	}
	else
	{
		print $handle <<EOT;
# 	name = value
# or
# 	name = "value"
#
# You can provide any number (including 0) of whitespaces before and after name and value. If you really want the whitespace in the value then use the second form and be happy;-)
EOT
	}
	print $handle <<EOT;
# It is also possible to set multiline strings with
# name <<ENDSTRING
# ...
# ENDSTRING
#	
# (just like in Perl but omitting the ;)
# You can use .=, +=, /= and *= instead of = as operators for concatenation of strings or pushing to arrays, addition, substraction, division and multiplication, respectively.
# The same holds likewise for .<<, +<<, /<< and *<< .
#
# The short names are just provided as a reference; they're only working as real command line parameters, not in this file!
#
# The lines starting with "=" are needed for parsers of the file (other than $config->{program} itself) and are informative to you, too.
# =param file (options) for program
# says for whom the file is and possibly some hints (options)
# =info INFO
# is the general program info (multiple lines, normally)
# =key NAME short S
# says that now comes stuff for the parameter NAME and its short form is S.
# =desc SOME_TEXT
# gives a description for the parameter.
#
# If you don't like/need all this bloated text, the you can strip all "#", "=" - started and empty lines and the result will still be a valid configuration file for $config->{program}.

EOT
	}
	print $handle '=param file ';
	my @opt = ();
	push(@opt,'eval') if $config->{eval};
	print $handle '('.join(',',@opt).') ' if @opt;
	print $handle 'for '.$config->{program}."\n";
	print $handle '=version '.$config->{version}."\n" if defined $config->{version}; 
	print $handle "\n";
	if(defined $config->{info} and !$bare)
	{
		my @info = split("\n",$config->{info});
		for(@info){ print $handle '=info '.$_."\n"; }
	}
	foreach my $k (sort keys %{$param})
	{
		unless(grep(/^$k$/, @omit))
		{
			#make line ending changeable...
			#or use proper system-independent line end
			#for now we just use \n - what may even work with active perl
			#
			unless($bare)
			{
				print $handle "\n=key $k",$short->{$k} ne '' ? " short $short->{$k}" : '',"\n";
				my @help = split("\n",${$par_help->{$k}});
				for(@help)
				{
					print $handle "=desc $_\n";
				}
			}
			my $values = $is_array->{$k} ? $param->{$k} : [ $param->{$k} ];
			$values = [ undef ] unless defined $values;
			my $first = 1;
			print $handle "\n" unless $bare;
			foreach my $v (@{$values})
			{
				my $preop = ($is_array->{$k} and not $first) ? '.' : '';
				if(defined $v and $v =~ /\n/)
				{
					my $end = 'EOT';
					my $num = '';
					$v =~ s/[\012\015]*\z/\n/;
					while($v =~ /(^|\n)$end$num(\n|$)/){ ++$num; }
					print $handle "$k $preop<<$end$num\n$v$end$num\n";
				}
				elsif($config->{eval} and not $raw)
				{
					#Are numbers strings? They could be... perl sorts it out anyway.
					print $handle defined $v ? "$k $preop= '$v'\n" : "\n$k = undef\n";
				}
				else
				{
					print $handle defined $v ? "$k $preop= \"$v\"\n" : "\n# $k is undefined\n";
				}
				$first = 0;
			}
		}
	}
}

sub SubSet
{
	my ($param, $def) = @_;
	my %ret;
	for(my $i = 0; $i <= $#{$def}; $i += 4)
	{
		$ret{$def->[$i]} = $param->{$def->[$i]};
	}
	return \%ret;
}

sub apply_op
{
	my ($parm, $op, $val, $is_array) = @_;
	unless($is_array)
	{
		# explicit again... I don't want to force quoting on the user
		# eval('$param{$key}'.$op.'$val');
		# my $ops = '.+\-*\/';
		if($$op eq '='){ $$parm = $$val;	}
		elsif($$op eq '.='){ $$parm .= $$val;	}
		elsif($$op eq '+='){ $$parm += $$val;	}
		elsif($$op eq '-='){ $$parm -= $$val;	}
		elsif($$op eq '*='){ $$parm *= $$val;	}
		elsif($$op eq '/='){ $$parm /= $$val;	}
		else{ die "[Param] Fatal internal error: Operator '$$op' is invalid, should not have slipped through parser!\n"; }
	}
	else
	{
		if($$op eq '=')
		{
			$$parm = [ $$val ];
		}
		elsif($$op eq '.=')
		{
			push(@{$$parm}, $$val);
		}
		else{ die "[Param] Fatal internal error: Operator '$$op' is invalid for array, should not have slipped through parser!\n"; }
	}
}

1;


__END__


=head1 NAME

Param - work with command line arguments and config files (yes, another one...)

=head1 SYNOPSIS

Just use the module, define your parameters.

	use Param;

	# the definitions in flat array
	# remember: parameters help / h	and config / I are predefined!
	my @params =
	(
	  'parm1', $default1,  'a', 'help text for scalar 1',
	  'parm2', $default2,  'b', 'help text for scalar 2',
	  'parmA', \@defaultA, 'A', 'help text for array A',
	  ...
	  'parmX', $defaultX,  'z', 'help text for last scalar'
	}

Then call the parser (of course you can define parameters directly as function arguments).

	$parm_ref = Param::Parse(@params);

	print "Value of parameter 'parm1': $parm_ref->{parm1}\n";
	print "Contents of array 'parmA': @{$parm_ref->{parmA}}\n";

or include some config for the parser

	my %config =
	(
	  'info' => 'program info text',
	  'version' => '1.2.3'
	  # possibly more configuration key/value pairs
	);
	$parm_ref = Param::Parse(\%config, @params);

or

	$parm_ref = Param::Parse(\%config,\@params); 

or

	$parm_ref = Param::Parse(\%config,\@params,\@commandline_args); 

=head1 DESCRIPTION

This module evolved from a simple loop for command line argument processing, that I thought was quicker written than to search for an already existing module for that. While just using Getopt I<may> have been a bit quicker, this wouldn't fully do what I want. It may fully do what I<you> want, though - so better check this before blaming me for wasting your time with installing Param.pm .

So, this is what I want and what therefore is provided by this module:

=over 4

=item * command line parameter processing

Process command line options/switches/parameters/... , be it short or long style, supporting clustering of short options. Interprets B<--parm=value> / B<-p=value> with the obvious effect; sets option to 1 (true) when just B<--option> / B<-o> is given.
Also, though somewhat counterintuitive but sort of a standard already and logical following the idea that "-" is true, B<++option> / B<+o> will set the value to 0 (false).
The form "--parm value" is not supported as no simple bullet-proof generic way to do that came across my mind. There is the fundamental problem of deciding if we have a parameter value or some other command line data like a file name to process. Since this problem still persists with the "=" used in assignment when one considers a file with a name like "--i_look_like_an_option", which is perfectly possible, Param also looks out for "--" as a final delimiter for the option part (like startx distinguishes between client options and server options). After "--" the Parse routine touches nothing.
A propos touch: The parameters are taken from the global @ARGV or the provided array (third parameter, see third example), which is shifted on processing. So everything interpreted by Param is deleted afterwards.

You can have scalars or arrays (reference) as values for your parameters. The array type is chosen when you provide an (anonymous) array reference as default value.

A nifty feature is the support of operators. Instead of B<--parm=value> you can do do B<--parm.=value>  to append something to the existing value. When B<-p> is the short form of B<--parm>, the same happens through B<-p.=value> or, saving one character, B<-p.value> .
So "B<--parm=a --parm.=b -p.c>" results in the value of B<parm> being "abc".

Besides ".=", these operators are available:

=over 4

=item * B<=>

Direct assignment.

=item * B<.=> or short B<.>

String concatenation for scalar parameters, pushing values for array parameters.

=item * B<+=> or short B<+>

Numeric addition to scalar value.

=item * B<-=> or short B<->

Numeric substraction to scalar value.

=item * B<*=> or short B<*>

Numeric multiplication of scalar value.

=item * B</=> or short B</>

Numeric division of scalar value.

=back

You can omit the B<=> on the command line when using special operators, but not in the configuration file.
There it is needed for parser safety. The operators extend to the multiline value parsing in config files, though (see the section on config file syntax).

See the B<lazy> configuration switch for a slightly modified command line syntax.

=item * automatic usage message creation

Based on the parameter definition Param automatically prints the expected usage/help message when the (predefined!) B<--help> / B<-h> was given, with the info string in advance when defined, and exits the program.

=item * configuration file parsing

The module also introduces a simple but flexible configuration file facility. Configuration means simply setting the same parameters that are available to the command line, with the added possibility to use perl syntax (expressions for eval(), the result being stored as value) when enabled in the program.
The second and last predefined parameter called "config" (short: I) is for giving a file (or several ones, see below!) to process I<before> the command line options are applied (what the user typed is always the last word). If none is given, some guessing may take place to find an appropriate file (see below).
When just B<-I> or B<--config> is given (no B<=>!), then a file with the current configuration ist written to STDOUT and the program exits (unless B<ignorefinals> is set).
When you give the lonely B<-I> or B<--config> once, you'll get a explained config file with comments and meta info, when you give the option twice, you get a condensed file with only the raw settings and a small header telling for which program it is.

Config files are parsed before applying command line arguments, so that the latter can override settings.

=back

As you probably have read, the usage consists mainly of calling the routine Param::Parse() with the parameter definitions in a flat array following an anonymous hash / a reference or just with hash/array references for the same stuff. The second variant includes the possibility to specify an array to use instead of @ARGV (per reference). 
The config hash configures the behaviour of Param.pm itself. No value is mandatory, but setting of the info string is highly recommened. All values are undefined (empty, 0, false) per default.


=head2 Config file syntax

In a configuration file only the long name of a parameter is used and values are explicit, example:

	mode = "something"

sets the parameter mode to the value something. Actually, 

	mode = something

would do the same. The whitespace around the operator (here: =) and before the line break is ignored.
You want to use (double) quotes when defining a value that begins or ends with whitespace:

	mode = " something spacy "

Consequently, you have to use quotes around the value if it starts with a quote character:

	mode = ""something with a quote"

The stuff between the first " and the last " will be used as value (you do not need to quote your quotes individually).
What is missing as up to now is the possibility to assign a multi-line string as value.
That you can do with a syntax very similar to a Perl construct:

	mode <<EOT
	a line
	another line
	bla
	EOT

You can replace the EOT with anything you like... even nothing - then, the next empty line marks the end.
Only rule is that the end marker should not appear in the multiline value itself (alone on a line).

You can also use the funny operators introduced for command line parameters by
replacing the B<=> with B<.=>, B<+=>, B<-=>, B</=>, B<*=> as you like.
That works for the B<<<EOT> multiline construct, too, just make it B<.<<EOT> for appending and likewise B<+<<EOT>, B<-<<EOT> , B</<<EOT>, B<*<<EOT> for the other operations.

Comments in the config file start with "#", special info tags for a parser (will be included in some future version of the Param package) start with "=":

	# this is a comment
	=param file for buntstift
	=version 0.25 (01.06.2007)
	
	=info wrapper around gnuplot for some quick shot plotting
	
	=key all-binary
	=desc specify binary stuff... one format for all files!
	all-binary = ""

Only the last line in this example is relevant to the program (named "buntstift" here). The rest is auxilliary information.

=head2 The config keys

=over 4

=item * info (some text)

A text to print out before the parameter info when giving help.

=item * program

name of the program/package, used in search for config file and for output of config files/messages; default is basename($0)

=item * version

Version string for your program; this activates the definition of a B<--version> switch to print it.

=item * verbose (0 / 1)

When not 0, be verbose about the work being done. There may be different verbosity levels in the future, but for now value 4 is no different from 1.

=item * multi (0 / 1)

Enable parsing of several configuration files one after another.

=item * confdir (some text)

Define a directory for searching of config files, overriding the internal default of "../etc" relative to the directory containing the program  (or the program dir itself if no etc there); explicitly given files are always searched in the current working directory first if the given form does not already specify an existing file completely.

=item * file (some text)

Define the default config file (s) (when using multiple ones: "file1,file2,file3,..."), overriding the internal default of confdir/programname.conf or (being more powerful if existing) confdir/programname.hostname.conf .
The .conf suffix will be added if needed.
A word on the hostname: This is a possibility to use one configuration directory for several hosts (p.ex. mounted over NFS). 

B<BEWARE:> Every smartness has its catch - when you do a `myprogram -I=my -I > my.conf` the newly created (and empty) file my.conf in the currect directoy is found and used as config file for reading, resulting in the program defaults being written regardless of what another existing my.conf would say!

=item * eval (0/ 1)

Activate the evaluation of parsed values from config files as code with eval(), this can be dangerous, but great fun, too;-)
One may even manipulate internals on Param.pm. I won't give a guide to that here... A simpler use is the possibility to specify a numeric constant like 4/3 properly or create a hash or array reference like that:

	array = [ 'val1', 'val2' ]
	hash = { 'key1', 'val1', 'key2', 'val2' }

But be careful with that... it may conflict with the use of "proper" array parameters and may look funny in help output.
You can always do

	$hash = eval "$param->{hash}";

in your program anyway.

=item * gimme (0 / 1)

Give help end exit if no command line arguments are there.

=item * lazy (0 / 1)

Switch to the "lazy" syntax that allows omitting the "-" and "--" in -option and --long-option (same with "+" and "++") most of the times but absolutely requires the separation of other stuff with "-+" (read this as PERL regex;-).

=item * ignorehelp (0 / 1)

ignore the help switch (when you don't want Param to exit() your program in any case), will vanish in version 3 in favour of nofinals

=item * nofinals (0 / 1)

Do not take final action on certain switches (help, version, config file printing). Normally (nofinals == 0), Param would exit your program on processing these.

=item * returnall (0 / 1)

makes Parse() return not only the parameter hash but also other stuff that is needed by the wrapper module Param::FileWorker

=item * notinfile (listref)

list with parameters (long names) not to include in config file printing, help and config are always left out as they make no sense in files (at least config could make sense if I allowed recursive file parsing in the future).

=item * nanny (0 / 1)

give parameter help and exit() when encountering unknown parameters in config files; otherwise just complain - this is disabled by nofinals/ignorehelp.

=item * fuzzy (0 / 1)

handle --non-existing-parameter by simply stopping parsing there (treat it as first file or such to be handled by your program), otherwise complain and dump help info + exit.

=item * nocomplain (0 / 1)

do not complain about missing config file or the like

=item * hidenonshort

hide parameters that don't have a short name in --help output (p.ex. multiline stuff that one doesn't want to see outside of config files)

=item * binmode (undef / :utf8 / :latin1 / ...)

Set the binmode (text encoding) for parsed files.

=back

=head1 AUTHOR

Thomas Orgis <thomas@orgis.org>

=cut


